// ---------------------------------------------------------------------------
/*
TODO:
Enable/Disable toolbar/menus based on state of editors
 */

#include <vcl.h>
#pragma hdrstop

#include "MainForm.h"

#include <string>
#include <algorithm>

// ---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TfrmMultiEditorMain *frmMultiEditorMain;

// ---------------------------------------------------------------------------
__fastcall TfrmMultiEditorMain::TfrmMultiEditorMain(TComponent* Owner) : TForm(Owner)
{
	// FFont = new TFont();
	// FFont->Name = "Courier New";
	// FFont->Size = 10;
	FIsClosing = false;
	FActiveEditor = 0;
	FTextEditors = new TObjectList(false);
	FUntitledFilesCount = 0;
}

__fastcall TfrmMultiEditorMain::~TfrmMultiEditorMain(void)
{
	delete FTextEditors;
	// delete FFont; // still in use
}

// ---------------------------------------------------------------------------
void __fastcall TfrmMultiEditorMain::FormCreate(TObject *Sender)
{
	// This *must* be done before DragAcceptFiles below. Apparently, it
	// changes something and the drag registration is erased afterwards.
	// Probably due to ReCreateWnd call.
	Position = poScreenCenter;

	// Allow files to be dropped on the app
	DragAcceptFiles(this->Handle, true);

	Width = 800;
	Height = 600;
	Caption = "Multi-File Editor";
	pnlEditor->Align = alClient;
	pgeEditors->Align = alClient;
	pgeEditors->ActivePageIndex = 0;

	ReadParams();

	if (FTextEditors->Count == 0)
		CreateNewEditor();
}

void TfrmMultiEditorMain::ReadParams(void)
{
	// TODO: Need to handle wildcards
	for (int i = 1; i <= ParamCount(); i++)
	{
		AnsiString s = ParamStr(i);
		CreateNewEditor(s);
	}
}

// ---------------------------------------------------------------------------

/*
This method gets called when a dropfiles message is received.
 */
void __fastcall TfrmMultiEditorMain::WMDropFiles(TWMDropFiles &Message)
{
	int num_files, filename_length;
	char file_name[MAX_PATH + 1];

	// Refer to the Win SDK documentation about the details
	// of DragQueryFile and the HDROP structure

	// Get the number of files that are being dropped on the app
	num_files = DragQueryFile((HDROP)Message.Drop, -1, NULL, 0);
	if (num_files < 1)
		return;

	// Retrieve each filename and open an editor for it
	for (int i = 0; i < num_files; i++)
	{
		// Get the length of the i'th filename
		filename_length = DragQueryFile((HDROP)Message.Drop, i, NULL, 0);

		// Retrieve the filename from the HDROP structure
		DragQueryFile((HDROP)Message.Drop, i, file_name, filename_length + 1);

		// Open file in a new editor
		CreateNewEditor(file_name);
	}
}

void __fastcall TfrmMultiEditorMain::FormDestroy(TObject *Sender)
{
	// App is no longer a drop target for files
	DragAcceptFiles(this->Handle, false);
}
// ---------------------------------------------------------------------------

void TfrmMultiEditorMain::CreateNewEditor(const AnsiString& Filename, bool load)
{
	bool load_file = load;
	AnsiString fname = Filename;
	if (Filename == "")
	{
		FUntitledFilesCount++;
		fname = Format("Untitled-%d.txt", ARRAYOFCONST((FUntitledFilesCount)));
		load_file = false;
	}
	else if (!FileExists(Filename))
	{
		fname = Filename;
		load_file = false;
	}

	TTextEditor *Editor = new TTextEditor(pgeEditors);
	// FActiveEditor.Font = FFont;
	// FActiveEditor.WordWrap = mnuOptionsWordWrap.Checked;
	// Editor->CaretPanel = barStatus->Panels->Items[1];
	FTextEditors->Add(Editor);
	pgeEditors->ActivePageIndex = FTextEditors->Count - 1;
	// Text := pgeEditors.Pages[pgeEditors.ActivePageIndex].Caption + ':' + IntToStr(FTextEditors.Count);

	Editor->EncodingType = etASCII;
	Editor->Filename = fname;
	FActiveEditor = Editor;

	if (load_file)
		Editor->OpenFile(fname);
}

// Returns true if the user canceled the operation, otherwise returns false
// This is the most "complicated" function in the program
bool TfrmMultiEditorMain::CloseEditor(int index)
{
	// If we are currently closing an editor, do nothing. You may ask "how can
	// this function be called again". Well, Windows is event driven, and if
	// the user tries to close all of the editors at once (hold the Ctrl-F4 key)
	// this function will be called again before it has finished from a previous
	// call. (It is re-entrant). So, to protect against that event, we just
	// only allow one at a time. BTW, this was discovered during stress-testing
	// when I had 50 editors opened and tried to close them all in rapid
	// succession. BOOM!
	if (FIsClosing)
		return false;

	// Set flag to protect from re-entry
	FIsClosing = true;

	// If we got a bogus index, just bail
	if ((index < 0) || (index >= pgeEditors->PageCount))
	{
		FIsClosing = false;
		return false;
	}

	// Get the editor from the list
	TTextEditor *Editor = (TTextEditor*)FTextEditors->Items[index];

	// If the file is "dirty", we need to ask the user what to do
	if (!Editor->CanClose())
	{
		// Format the warning message
		AnsiString file = ExtractFileName(Editor->Filename);
		AnsiString msg = Format("Save changes to %s?", ARRAYOFCONST((file)));
		int key = MessageDlg(msg, mtWarning, TMsgDlgButtons() << mbYes << mbNo << mbCancel, 0);

		// If they canceled, do nothing
		if (key == mrCancel)
		{
			FIsClosing = false;
			return true;
		}

		// If they answered yes
		if (key == mrYes)
		{
			// If the file hasn't been saved yet (no real filename),
			// then show a File Save dialog
			if (Editor->RequiresInitialSaveDialog)
			{
				// If the user pressed OK, they chose a name and want to save
				if (dlgFileSave->Execute())
					Editor->SaveFileAs(dlgFileSave->FileName, etASCII, saDefault);
				else // user canceled
				{
					FIsClosing = false;
					return true;
				}
			}
			// the file has been saved previously, so we just save under
			// that same name
			else
				Editor->SaveFile();

			FIsClosing = false;
		}
	}

	// The file has been saved, if necessary, at this point so now
	// it's time to remove the editor from the pages

	int NewIndex, curr_index;
	bool keep_focus = false;

	// If we're closing a tab that isn't focused, the focused tab should remain
	// focused after the tab is removed.
	if (index != pgeEditors->ActivePageIndex)
		keep_focus = true;

	curr_index = pgeEditors->ActivePageIndex;

	// Remove the editor from the list
	FTextEditors->Remove(Editor);
	Application->ProcessMessages(); // Yield

	// If we deleted the last one, make the penultimate one the focus (maybe)
	// otherwise, well make the one that was next to the removed one the focus
	if (index < FTextEditors->Count)
		NewIndex = index;
	else
		NewIndex = FTextEditors->Count - 1;

	// Delete the editor
	Editor->Free();
	Application->ProcessMessages();

	// If there are no more editors open, don't focus anything
	if (NewIndex == -1)
	{
		FIsClosing = false;
		return false;
	}

	// If we removed a tab that came before the focused tab, need to
	// update our position (invalidated iterator pattern)
	if (curr_index > index)
		curr_index--;

	// Set active editor to the one that gets the focus
	if (!keep_focus)
	{
		pgeEditors->ActivePageIndex = NewIndex;
		FActiveEditor = (TTextEditor*)(FTextEditors->Items[NewIndex]);
	}
	else
	{
		if (curr_index >= pgeEditors->PageCount)
			curr_index = pgeEditors->PageCount - 1;

		pgeEditors->ActivePageIndex = curr_index;
		FActiveEditor = (TTextEditor*)(FTextEditors->Items[curr_index]);
	}

	Application->ProcessMessages();
	FIsClosing = false;
	return false;
}

void __fastcall TfrmMultiEditorMain::pgeEditorsMouseDown(TObject *Sender, TMouseButton Button,
	TShiftState Shift, int X, int Y)
{
	if (Button == mbMiddle)
	{
		int index = pgeEditors->IndexOfTabAt(X, Y);
		CloseEditor(index);
	}
	else
		pgeEditors->BeginDrag(false);
}
// ---------------------------------------------------------------------------

void __fastcall TfrmMultiEditorMain::pgeEditorsChange(TObject *Sender)
{
	if (FTextEditors->Count > 0)
		FActiveEditor = (TTextEditor*)(FTextEditors->Items[pgeEditors->ActivePageIndex]);
}
// ---------------------------------------------------------------------------

void __fastcall TfrmMultiEditorMain::actFileNewExecute(TObject *Sender)
{
	CreateNewEditor();
}
// ---------------------------------------------------------------------------

void __fastcall TfrmMultiEditorMain::actFileExitHint(AnsiString &HintStr, bool &CanShow)
{
	Close();
}
// ---------------------------------------------------------------------------

void __fastcall TfrmMultiEditorMain::actFileCloseExecute(TObject *Sender)
{
	CloseEditor(pgeEditors->ActivePageIndex);
}
// ---------------------------------------------------------------------------

void __fastcall TfrmMultiEditorMain::actEditSelectFontExecute(TObject *Sender)
{
	if (FFont)
		dlgFont->Font = FFont;
	if (dlgFont->Execute())
	{
		FFont = dlgFont->Font;
		for (int i = 0; i < FTextEditors->Count; i++)
		{
			bool dirty = (((TTextEditor*)FTextEditors->Items[i]))->Dirty;
			(((TTextEditor*)FTextEditors->Items[i]))->Font = FFont;
			if (!dirty)
				(((TTextEditor*)FTextEditors->Items[i]))->ClearDirtyFlag();

		}
	}
}
// ---------------------------------------------------------------------------

void __fastcall TfrmMultiEditorMain::actFileOpenBeforeExecute(TObject *Sender)
{
	// Get existing options
	TOpenOptions oo = actFileOpen->Dialog->Options;

	// Add additional options
	oo << ofAllowMultiSelect;
	actFileOpen->Dialog->Options = oo;
}
// ---------------------------------------------------------------------------

void __fastcall TfrmMultiEditorMain::actFileOpenAccept(TObject *Sender)
{
	for (int i = 0; i < actFileOpen->Dialog->Files->Count; i++)
	{
		AnsiString fname = actFileOpen->Dialog->Files->Strings[i];
		CreateNewEditor(fname);
	}
}
// ---------------------------------------------------------------------------

void __fastcall TfrmMultiEditorMain::actFileSaveExecute(TObject *Sender)
{
	TTextEditor *editor = GetActiveEditor();
	if (editor->RequiresInitialSaveDialog)
	{
		if (dlgFileSave->Execute())
			editor->SaveFileAs(dlgFileSave->FileName, etASCII, saDefault);
	}
	else
		editor->SaveFile();
}
// ---------------------------------------------------------------------------

TTextEditor *TfrmMultiEditorMain::GetActiveEditor(void)
{
	if (FTextEditors->Count > 0)
	{
		FActiveEditor = (TTextEditor*)(FTextEditors->Items[pgeEditors->ActivePageIndex]);
		return FActiveEditor;
	}
	else
		return 0;
}

void __fastcall TfrmMultiEditorMain::actFileSaveAsAccept(TObject *Sender)
{
	TTextEditor *editor = GetActiveEditor();
	editor->SaveFileAs(actFileSaveAs->Dialog->FileName, etASCII, saDefault);
}
// ---------------------------------------------------------------------------

void __fastcall TfrmMultiEditorMain::actHelpAboutExecute(TObject *Sender)
{
	// frmAbout->ShowModal();
	ShowMessage("About box will go here");
}
// ---------------------------------------------------------------------------

void __fastcall TfrmMultiEditorMain::FormCloseQuery(TObject *Sender, bool &CanClose)
{
	CanClose = !CloseAllEditors();
}
// ---------------------------------------------------------------------------

void __fastcall TfrmMultiEditorMain::actFileCloseAllExecute(TObject *Sender)
{
	CloseAllEditors();
}
// ---------------------------------------------------------------------------

bool TfrmMultiEditorMain::CloseAllEditors(void)
{
	int count = pgeEditors->PageCount;
	for (int i = 0; i < count; i++)
	{
		bool canceled = CloseEditor(pgeEditors->ActivePageIndex);
		if (canceled)
			return true;
	}
	return false;
}
