//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "TextEditor.h"
#include <string>


//---------------------------------------------------------------------------

#pragma package(smart_init)

const char* DEFAULT_FILENAME = "Untitled.txt";

TTextEditor::TTextEditor(TPageControl *PageControl) :
#ifdef USE_RICHEDIT
TRichEdit(PageControl)
#else
TMemo(PageControl)
#endif
{
	/* Get double buffering working with TRichEdit
	http://qc.codegear.com/wc/qcmain.aspx?d=41125
	http://www.experts-exchange.com/Programming/Languages/Pascal/Delphi/Q_23599288.html

	int Mask = SendMessage(this->Handle,EM_SETEVENTMASK,0,0);
	SendMessage(this->Handle,WM_SETREDRAW,false,0);
	//{DO what ever you do to the richedit here}
	SendMessage(this->Handle,WM_SETREDRAW,true,0);
	InvalidateRect(this->Handle,0,true);
	SendMessage(this->Handle,EM_SETEVENTMASK,0,Mask);
	Beep();
	*/


	FFilename = DEFAULT_FILENAME;
	FTabSheet = new TTabSheet(PageControl);
	FTabSheet->Caption = FFilename;
	FTabSheet->PageControl = PageControl;

/* From the help:
	 Some controls, such as TRichEdit, can't paint themselves into a bitmap.
	 For such controls, DoubleBuffered must be set to false.
*/
#ifndef USE_RICHEDIT
	this->DoubleBuffered = true;
#endif

	this->Parent = FTabSheet;
	this->Align = alClient;
	this->Font->Name = "Courier New";
	this->Font->Size = 10;

#ifdef USE_RICHEDIT
	this->WordWrap = true;
	this->ScrollBars = ssBoth;
#else
	this->WordWrap = true;
	this->ScrollBars = ssVertical; // ssBoth for TMemo if word wrap is false
#endif

	this->OnChange = Changed;
	this->OnClick = EditorClick;
	this->OnKeyDown = EditorKeyDown;
	this->WantTabs = true;

#ifdef USE_RICHEDIT
	this->HideScrollBars = true;
#endif

	this->HideSelection = false;
	this->TabStop = false;

	FInUse = false;
	FRequiresInitialSaveDialog = true;

		// Work-around for bogus Change events (see Changed function below)
	FCurrentHWND = this->Handle;
}

__fastcall TTextEditor::~TTextEditor(void)
{
	FTabSheet->PageControl = NULL;
}

void TTextEditor::DisplayFilename(void)
{
	AnsiString Str;
	if (FDirty)
		Str = "*";
	else
		Str = "";

	FTabSheet->Caption = ExtractFileName(FFilename + Str);
	Hint = FFilename + "   (Click middle mouse button on Tab to close file)";
	ShowHint = False;
}

void TTextEditor::NewDocument(void)
{
	this->Clear();
	FFilename = "Untitled.txt";
	ClearDirtyFlag();
}

void TTextEditor::OpenFile(const AnsiString& Filename, TEncodingType EncodingType)
{
	#if 0
	TFileStream *FileStream;
	if (FileIsUnicode(Filename))
	{
		FileStream = new TFileStream(Filename, fmOpenRead);
		this->Lines->LoadFromStream(FileStream);
		FileStream->Free();
		FEncodingType = etUTF16;
	}
	else
	{
		FEncodingType = etASCII;
		this->Lines->LoadFromFile(Filename);
	}
	#endif

	LoadFromFile(Filename);

	FRequiresInitialSaveDialog = false;
	FFilename = Filename;
	FInUse = true;
	ClearDirtyFlag();
	DisplayFilename();
}

void __fastcall TTextEditor::Changed(TObject *Sender)
{
/* HACK?

	It turns out that when the page control changes from single line tabs to
	multiline tabs, the underlying Window handle (HWND) is recreated from scratch.
	This has the nice "feature" that causes a Changed message sent to all
	components on the tab sheet. This is a problem because *nothing* has actually
	changed in the editor. So, this code kind of works around that. Essentially:

	We get our HWND when the editor is created. If we get a Changed message and
	the HWND has changed, we assume that the reason for the message was because
	of the HWND recreation. Set the CurrentHWND to the new one so that the *next*
	time we get a Changed message, we will know that it is a *real* change in
	the editor itself.

	Haven't thought through this for too long so I don't know if there is some
	other subtle area where we might get a change notification when we really
	shouldn't.
*/

	HWND h = this->Handle;
	//ShowMessage(Format("%d", ARRAYOFCONST((Integer(h)))));

	if (FCurrentHWND == h) // This was a real change in the editor, mark dirty
		SetDirtyFlag();
	else                   // Extraneous change, don't mark the editor as dirty
		FCurrentHWND = h;
}

void TTextEditor::ReloadFromDisk(void)
{
	OpenFile(FFilename);
}

void TTextEditor::SaveFile(const AnsiString& Filename, bool UpdateTabCaption, bool ClearDirty)
{
	TFileStream *FileStream;
	FileStream = 0;
	if ((FEncodingType == etASCII) || (FEncodingType == etUTF16))
	{
		try
		{
			FileStream = new TFileStream(Filename, fmCreate);
			this->Lines->SaveToStream(FileStream);
			FRequiresInitialSaveDialog = false;
			if (ClearDirty)
				ClearDirtyFlag();
		if (UpdateTabCaption)
				DisplayFilename();
		}
		__finally
		{
			FileStream->Free();
		}
	}
}

void TTextEditor::SaveFile(void)
{
	SaveFile(FFilename);
}

void TTextEditor::SaveToTemp(const AnsiString& Filename)
{
	bool dirt = FDirty; // preserve the dirty flag

#ifdef USE_RICHEDIT
	PlainText = true; // RTF will be saved as plain text 
#endif

	SaveFile(Filename, false, false);

#ifdef USE_RICHEDIT
	PlainText = false;
#endif

	FDirty = dirt;
}

void TTextEditor::SaveFileAs(const AnsiString& Filename, TEncodingType EncodingType, TSignatureAction SignatureAction)
{
	FEncodingType = EncodingType;
	FSignatureAction = SignatureAction;
	SaveFile(Filename);
	FFilename = Filename;
	DisplayFilename();
}

void TTextEditor::SetDirectory(const AnsiString& Value)
{
	FDirectory = Value;
}

void TTextEditor::SetDirty(bool Value)
{
	FDirty = Value;
}

void TTextEditor::SetRequiresInitialSaveDialog(bool value)
{
  FRequiresInitialSaveDialog = value;
}

void TTextEditor::SetDirtyFlag(void)
{
  if (FDirty)
		return;

  FInUse = true;
  FDirty = true;
  DisplayFilename();
}

void TTextEditor::SetEncodingType(TEncodingType Value)
{
  FEncodingType = Value;
}

void TTextEditor::SetFilename(const AnsiString& Value)
{
	if (FFilename == Value)
		return;
		
	FFilename = Value;
	DisplayFilename();
}

void TTextEditor::SetSignatureAction(TSignatureAction Value)
{
	FSignatureAction = Value;
}

void TTextEditor::LoadFromFile(const AnsiString& FileName)
{
	try
	{
		this->Lines->LoadFromFile(FileName);
	}
	__finally
	{
  	// no longer used
	}
}

void TTextEditor::LoadFromStream(TStream* Stream)
{
	this->Lines->BeginUpdate();
	try
	{
		int Size = int(Stream->Size - Stream->Position);
		AnsiString S;
		S.SetLength(Size);
		Stream->Read(S.c_str(), Size);

		String s(S.c_str());
		this->Lines->SetText(s.c_str());
	}
	__finally
	{
		// no longer used
	}
	this->Lines->EndUpdate();
}

void TTextEditor::ClearDirtyFlag(void)
{
	if (!FDirty)
		return;

	FDirty = false;
	DisplayFilename();
}


bool TTextEditor::CanClose(void)
{
	return !Dirty;
}

void __fastcall TTextEditor::EditorClick(TObject *)
{
	UpdateCaretPanel();
}

void __fastcall TTextEditor::EditorKeyDown(TObject *, Word& Key, TShiftState Shift)
{
	UpdateCaretPanel();
}

void TTextEditor::SetCaretPanel(TStatusPanel *Value)
{
	FCaretPanel = Value;
}

void TTextEditor::UpdateCaretPanel(void)
{
	//FCaretPanel->Text = Format("%d: %d", ARRAYOFCONST((this->SelStart->y, this->SelStart.x)));
	//FCaretPanel->Text = Format("%d: %d", ARRAYOFCONST((this->SelStart, this->SelLength)));
}


