//---------------------------------------------------------------------------

#ifndef TextEditorH
#define TextEditorH

enum TFileEncoding {feASCII, feUTF16, feUTF8};
enum TEncodingType {etDefault, etUTF16, etUTF8, etASCII};
enum TSignatureAction {saDefault, saPreserve, saAdd, saRemove};

#define USE_RICHEDITx

class TTextEditor : public
#ifdef USE_RICHEDIT
TRichEdit
#else
TMemo
#endif
{
	private:
		TTabSheet *FTabSheet;
		bool FDirty;
		AnsiString FFilename;
		AnsiString FDirectory;
		bool FInUse;
		bool FRequiresInitialSaveDialog;
		TEncodingType FEncodingType;
		TSignatureAction FSignatureAction;
		TStatusPanel *FCaretPanel;
		HWND FCurrentHWND;

	protected:
		void LoadFromFile(const AnsiString& FileName);
		void LoadFromStream(TStream *Stream);

		void __fastcall EditorClick(TObject *Sender);
		void __fastcall EditorKeyDown(TObject *Sender, Word& Key, TShiftState Shift);
		void __fastcall Changed(TObject *Sender);

		void UpdateCaretPanel();

		void SaveFile(const AnsiString& Filename, bool UpdateTabCaption = true, bool ClearDirty = true);
		void SetDirectory(const AnsiString& Value);
		void SetDirty(bool value);
		void SetRequiresInitialSaveDialog(bool value);
		void SetFilename(const AnsiString& Value);
		void SetEncodingType(TEncodingType Value);
		void SetSignatureAction(TSignatureAction Value);
		void SetCaretPanel(TStatusPanel* Value);
		
	public:
		__property bool RequiresInitialSaveDialog = {read=FRequiresInitialSaveDialog, write=SetRequiresInitialSaveDialog};
		__property bool Dirty = {read=FDirty, write=SetDirty};
		__property AnsiString Filename = {read=FFilename, write=SetFilename};
		__property AnsiString Directory = {read=FDirectory, write=SetDirectory};
		__property bool InUse = {read=FInUse};
		__property TSignatureAction SignatureAction = {read=FSignatureAction, write=SetSignatureAction};
		__property TEncodingType EncodingType = {read=FEncodingType, write=SetEncodingType};
		__property TStatusPanel *CaretPanel = {read=FCaretPanel, write=SetCaretPanel};

		TTextEditor(TPageControl *PageControl);
		__fastcall virtual ~TTextEditor(void);

		void DisplayFilename(void);
		void OpenFile(const AnsiString& Filename, TEncodingType EncodingType = etDefault);
		void SaveFile(void);
		void SaveFileAs(const AnsiString& Filename, TEncodingType EncodingType, TSignatureAction SignatureAction);
		void ClearDirtyFlag(void);
		void SetDirtyFlag(void);
		void NewDocument(void);
		void ReloadFromDisk(void);
		void SaveToTemp(const AnsiString& Filename);
		bool CanClose(void);
};



//---------------------------------------------------------------------------
#endif
